function [dualUpper,L,U] = derivedualBounds(H,c,A,b,E,f,ops,parametricDomain)

if isempty(A)
    dualUpper = [];
    L = [];
    U = [];
    return
end

n = length(c);
m = length(b);

x = sdpvar(n,1);
Lambda = sdpvar(m,1);

F = A*x < b;

% Start by computing primal lower bounds
if nargin < 7
    ops = sdpsettings('verbose',0);
end
ops2 = ops;
ops2.verbose = max(0,ops.verbose-1);;
all_bounded = 1;
if ops.verbose
    disp(['Deriving ' num2str(length(x)) ' primal bounds (required for dual bounds)']);
end

z = recover(depends(c));
xz = [x;z];
nz = length(z);
nTOT = n + length(z);
sol = solvesdp([F,parametricDomain,H*x + c + A'*Lambda==0,Lambda>0],[xz;-xz],ops2);
for i = 1:nTOT
    selectsolution(i);
    L(i,1) = double(xz(i));
    selectsolution(nTOT+i);
    U(i,1) = double(xz(i));
end

c0C = full(getbase(c));
c0 = c0C(:,1);
C = c0C(:,2:end);

b0B = full(getbase(b));
b0 = b0B(:,1);
B = b0B(:,2:end);


% Lift
XZ = sdpvar(nTOT);
constraints = [Lambda > 0,parametricDomain,F,L<xz<U];
% 
for i = 1:nTOT
    for j = i:nTOT
        constraints = [constraints, XZ(i,j)-L(i)*xz(j)-L(i)*xz(j)+L(i)*L(j)>0];
        constraints = [constraints, XZ(i,j)-U(i)*xz(j)-U(i)*xz(j)+U(i)*U(j)>0];
        constraints = [constraints, U(i)*xz(j)-U(i)*L(j)-XZ(i,j)+xz(i)*L(j)>0];
        constraints = [constraints, U(j)*xz(i)-U(j)*L(i)-XZ(i,j)+xz(j)*L(i)>0];
    end
end

if ops.verbose
    disp(['Deriving ' num2str(length(Lambda)) ' dual bounds']);
end
ops2.verbose = 0;
solvesdp(constraints,-Lambda,ops2);
for i = 1:m
    selectsolution(i);  
    dualUpper(i,1) = double(Lambda(i));
end

